﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;

namespace Hamlet
{
    [Flags]
    public enum EdgeEnds
    {
        Start = 1,
        End = 2,
        Both = 3
    }

    /// <summary>
    /// Represents a generic graph.
    /// </summary>
    /// <typeparam name="V">This is the type of data contained in the graph's vertices. It must implement IEquatable<V>.</typeparam>
    /// <typeparam name="E">This is the type of data associated to the graph's edges. It must implement IComparable<E>.</typeparam>
    public class Graph<V, E>
        where V : IEquatable<V> 
        where E : IEquatable<E>
    {
        /// <summary>
        /// Represents a vertex.
        /// </summary>
        [DebuggerDisplay("{ToString()}")]
        public class Vertex : IEquatable<Vertex>
        {
            /// <summary>
            /// Gets the index of this vertex in the vertex collection.
            /// </summary>
            public Int32 Index { get; internal set; }
            /// <summary>
            /// Gets or sets the data contained in this vertex.
            /// </summary>
            public V Data { get; set; }

            public override string ToString()
            {
                return '{' + this.Data.ToString() + ", " + this.Index + '}';
            }

            public bool Equals(Vertex other)
            {
                if (other == null)
                    return false;
                return (this.Index == other.Index) && this.Data.Equals(other.Data);
            }
        }

        /// <summary>
        /// Represents an edge.
        /// </summary>
        [DebuggerDisplay("{ToString()}")]
        public class Edge : IEquatable<Edge>
        {
            /// <summary>
            /// Gets or sets the starting vertex of this edge.
            /// </summary>
            public Vertex Start { get; set; }
            /// <summary>
            /// Gets or sets the ending vertex of this edge.
            /// </summary>
            public Vertex End { get; set; }
            /// <summary>
            /// Gets or sets the data associated to this edge.
            /// </summary>
            public E Data { get; set; }

            public override string ToString()
            {
                return String.Format("{0} : {1} -> {2}", this.Data.ToString(), this.Start.ToString(), this.End.ToString());
            }

            public bool Equals(Edge other)
            {
                if (other == null)
                    return false;
                return (this.Start == other.Start) && (this.End == other.End) && this.Data.Equals(other.Data);
            }
        }

        public class StronglyConnectedComponent : ReadOnlyCollection<Vertex>
        {
            public StronglyConnectedComponent(IList<Vertex> source) : base(source) { }
        }

        protected List<Vertex> vertices;
        protected List<Edge> edges;
        protected IStronglyConnectedComponentsProvider<V, E> defaultSccProvider;

        /// <summary>
        /// Rerturns the collection of all data exposed by vertices, sorted in the same order as the vertices themselves.
        /// </summary>
        public IEnumerable<V> VerticesData
        {
            get
            {
                foreach (Vertex v in this.Vertices)
                    yield return v.Data;
            }
        }

        /// <summary>
        /// Returns the collection of all data exposed by edges.
        /// </summary>
        public IEnumerable<E> EdgesData
        {
            get
            {
                foreach (Edge e in this.Edges)
                    yield return e.Data;
            }
        }

        /// <summary>
        /// Gets the order of the graph, i.e. the number of its vertices.
        /// </summary>
        public Int32 Order
        {
            get { return this.vertices.Count; }
        }

        /// <summary>
        /// Gets the size of the graph, i.e. the numbers of its edges.
        /// </summary>
        public Int32 Size
        {
            get { return this.edges.Count; }
        }

        /// <summary>
        /// Determins if this graph is trivial, i.e. it has only 1 vertex.
        /// </summary>
        public Boolean IsTrivial
        {
            get { return this.vertices.Count == 1; }
        }
        
        /// <summary>
        /// Gets or sets a value indicating whether the graph is directed.
        /// </summary>
        public Boolean IsDirected { get; set; }
        /// <summary>
        /// Exposes the collection of all vertices.
        /// </summary>
        public ReadOnlyCollection<Vertex> Vertices { get; protected set; }
        /// <summary>
        /// Exposes the collection of all edges.
        /// </summary>
        public ReadOnlyCollection<Edge> Edges { get; protected set; }

        /// <summary>
        /// Initializes a new empty graph.
        /// </summary>
        public Graph()
        {
            vertices = new List<Vertex>();
            edges = new List<Edge>();
            this.Vertices = vertices.AsReadOnly();
            this.Edges = edges.AsReadOnly();
        }

        /// <summary>
        /// Adds a new vertex to the graph
        /// </summary>
        /// <param name="data">The data contained the vertex.</param>
        public void AddVertex(V data)
        {
            Vertex vertex = new Vertex() { Data = data, Index = vertices.Count };
            vertices.Add(vertex);
        }

        /// <summary>
        /// Removes a vertex from the graph, provided its content.
        /// </summary>
        /// <param name="data">Indicates the data a vertex must posses to be deleted from the graph.</param>
        public void RemoveVertex(V data)
        {
            this.RemoveVertices((from v in vertices where v.Data.Equals(data) select v).ToArray());
        }

        /// <summary>
        /// Removes a vertex from the graph, provided its index.
        /// </summary>
        /// <param name="index">Indicates the index of the vertex being removed.</param>
        public void RemoveVertex(Int32 index)
        {
            if (index < 0 || index >= vertices.Count)
                throw new IndexOutOfRangeException();
            this.RemoveVertices(new Vertex[] { vertices[index] });
        }

        private void RemoveVertices(IEnumerable<Vertex> matches)
        {
            foreach (var vertex in matches)
            {
                vertices.Remove(vertex);
                var adjacentEdges = this.FindEdges(vertex, EdgeEnds.Both).ToArray();
                foreach (var edge in adjacentEdges)
                    edges.Remove(edge);
            }
            this.RefreshVertexIndices();
        }

        private void RefreshVertexIndices()
        {
            for (Int32 i = 0; i < vertices.Count; i++)
                vertices[i].Index = i;
        }

        /// <summary>
        /// Creates a new edge linking two existing vertices.
        /// </summary>
        /// <param name="vertexIndex1">Index of the starting vertex.</param>
        /// <param name="vertexIndex2">Index of the ending vertex.</param>
        /// <param name="data">Data associated to the edge.</param>
        public void AddEdge(Int32 vertexIndex1, Int32 vertexIndex2, E data)
        {
            if ((vertexIndex1 < 0 || vertexIndex1 >= vertices.Count) ||
                (vertexIndex2 < 0 || vertexIndex2 >= vertices.Count))
                throw new IndexOutOfRangeException();

            Vertex v1 = vertices[vertexIndex1];
            Vertex v2 = vertices[vertexIndex2];
            Edge e = new Edge() { Start = v1, End = v2, Data = data };
            edges.Add(e);
        }

        /// <summary>
        /// Creates a new edge linking two existing vertices, provided their content.
        /// </summary>
        /// <param name="vertexData1">Data used to determine which vertex is used as starting vertex.</param>
        /// <param name="vertexData2">Data used to determine which vertex is used as ending vertex.</param>
        /// <param name="data">Data associated to the edge.</param>
        public void AddEdge(V vertexData1, V vertexData2, E data)
        {
            try
            {
                Int32 index1 = vertices.Single(v => v.Data.Equals(vertexData1)).Index;
                Int32 index2 = vertices.Single(v => v.Data.Equals(vertexData2)).Index;
                this.AddEdge(index1, index2, data);
            }
            catch
            {
                throw new ArgumentException("When specifing vertices with their data content, you cannot pass data shared among more vertices.");
            }
        }

        /// <summary>
        /// Removes an existing edge between two vertices. If there are more than one edge in between the vertices,
        /// the method will remove all of them. If the graph is directed only edges going straight from the first
        /// to the second will be considered.
        /// </summary>
        /// <param name="vertexIndex1">Index of the starting vertex.</param>
        /// <param name="vertexIndex2">Index of the ending vertex.</param>
        public void RemoveEdge(Int32 vertexIndex1, Int32 vertexIndex2)
        {
            if ((vertexIndex1 < 0 || vertexIndex1 >= vertices.Count) ||
                (vertexIndex2 < 0 || vertexIndex2 >= vertices.Count))
                throw new IndexOutOfRangeException();

            IEnumerable<Edge> edges1, edges2;

            if (this.IsDirected)
            {
                edges1 = this.FindEdges(vertices[vertexIndex1], EdgeEnds.Start);
                edges2 = this.FindEdges(vertices[vertexIndex2], EdgeEnds.End);
            }
            else
            {
                edges1 = this.FindEdges(vertices[vertexIndex1], EdgeEnds.Both);
                edges2 = this.FindEdges(vertices[vertexIndex2], EdgeEnds.Both);
            }

            var matches = edges1.Intersect(edges2).ToArray();
            foreach (var edge in matches)
                edges.Remove(edge);
        }

        /// <summary>
        /// Removes an existing edge between two vertices. If there are more than one edge in between the vertices,
        /// the method will remove all of them. If the graph is directed only edges going straight from the first
        /// to the second will be considered. 
        /// </summary>
        /// <param name="vertexData1">Data used to determine which vertex is used as starting vertex.</param>
        /// <param name="vertexData2">Data used to determine which vertex is used as ending vertex.</param>
        public void RemoveEdge(V vertexData1, V vertexData2)
        {
            try
            {
                Int32 index1 = vertices.Single(v => v.Data.Equals(vertexData1)).Index;
                Int32 index2 = vertices.Single(v => v.Data.Equals(vertexData2)).Index;
                this.RemoveEdge(index1, index2);
            }
            catch
            {
                throw new ArgumentException("When specifing vertices with their data content, you cannot pass data shared among more vertices.");
            }
        }

        /// <summary>
        /// Find all the edges starting or ending with the specified vertex.
        /// </summary>
        /// <param name="ends">Flags indicating whether to considering incoming or outcoming edges, or both.</param>
        public IEnumerable<Edge> FindEdges(Vertex vertex, EdgeEnds ends)
        {
            foreach (Edge e in edges)
            {
                if ( (e.Start.Equals(vertex) && ends.HasFlag(EdgeEnds.Start)) || (e.End.Equals(vertex) && ends.HasFlag(EdgeEnds.End)) )
                    yield return e;
            }
        }

        /// <summary>
        /// Find all the edges starting at vertex1 and ending at vertex2.
        /// </summary>
        public IEnumerable<Edge> FindEdges(Vertex vertex1, Vertex vertex2)
        {
            foreach (Edge e in edges)
                if ( (e.Start.Equals(vertex1) && e.End.Equals(vertex2)) || (!this.IsDirected && e.End.Equals(vertex1) && e.Start.Equals(vertex2)) )
                    yield return e;
        }

        /// <summary>
        /// Check whether two vertices are adjacent.
        /// </summary>
        public Boolean AreVerticesAdjacent(Vertex vertex1, Vertex vertex2)
        {
            foreach(Edge e in edges)
                if ( (e.Start.Equals(vertex1) && e.End.Equals(vertex2)) || (e.Start.Equals(vertex2) && e.End.Equals(vertex1)) )
                    return true;
            return false;
        }

        /// <summary>
        /// Check whether two edges are adjacent.
        /// </summary>
        public Boolean AreEdgesAdjacent(Edge edge1, Edge edge2)
        {
            return (edge1.Start.Equals(edge2.End)) || (edge1.End.Equals(edge2.Start));
        }

        /// <summary>
        /// Check whether this graph is a multigraph.
        /// </summary>
        public Boolean IsMultigraph()
        {
            for (Int32 i = 0; i < edges.Count; i++)
                for (Int32 j = i + 1; j < edges.Count; j++)
                    if (edges[i].End.Equals(edges[j].End))
                        return true;
            return false;
        }

        /// <summary>
        /// Calculates and returns the adjacency matrix for this graph.
        /// </summary>
        public Byte[,] GetAdjacencyMatrix()
        {
            Byte[,] matrix = new Byte[this.Order, this.Order];

            foreach (Edge e in edges)
            {
                matrix[e.Start.Index, e.End.Index] = 1;
                if (this.IsDirected)
                    matrix[e.End.Index, e.Start.Index] = 1;
            }

            return matrix;
        }

        /// <summary>
        /// Retrieves the neighbourhood set of a given vertex.
        /// </summary>
        public IEnumerable<Vertex> GetNeighbourhoodSet(Vertex vertex)
        {
            if (this.IsDirected)
                return this.FindEdges(vertex, EdgeEnds.Start)
                           .Select(e => e.End)
                           .Distinct()
                           .Where(v => !v.Equals(vertex));

            return this.FindEdges(vertex, EdgeEnds.Both)
                       .SelectMany(e => new Vertex[] { e.Start, e.End })
                       .Distinct()
                       .Where(v => !v.Equals(vertex));
        }

        /// <summary>
        /// Gets the degree (i.e. the cardinality of neighbourhood set) of a given vertex.
        /// </summary>
        public Int32 GetDegree(Vertex vertex)
        {
            return this.GetNeighbourhoodSet(vertex).Count();
        }

        /// <summary>
        /// Check whether this graph is subgraph of another graph.
        /// </summary>
        public Boolean IsSubgraphOf(Graph<V, E> superGraph)
        {
            return this.vertices.All(v => superGraph.vertices.Contains(v)) &&
                   this.edges.All(e => superGraph.edges.Contains(e));
        }

        /// <summary>
        /// Check whether this graph spans another graph.
        /// </summary>
        public Boolean Spans(Graph<V, E> otherGraph)
        {
            return this.vertices.SequenceEqual(otherGraph.vertices);
        }

        /// <summary>
        /// Check whether this graph is complete, i.e. all vertices are adjacent.
        /// </summary>
        public Boolean IsComplete()
        {
            for (Int32 i = 0; i < vertices.Count; i++)
                for (Int32 j = i + 1; j < vertices.Count; j++)
                    if (!this.AreVerticesAdjacent(vertices[i], vertices[j]))
                        return false;
            return true;
        }

        /// <summary>
        /// Check whether this graph is regular, i.e. the degree of all vertices is the same.
        /// </summary>
        public Boolean IsRegular()
        {
            if (vertices.Count == 0)
                return true;

            Int32 degree = this.GetDegree(vertices[0]);
            return vertices.All(v => this.GetDegree(v) == degree);
        }

        /// <summary>
        /// Check whether this graph is bipartite, given a function to classify its vertices in two sets (X and Y).
        /// </summary>
        /// <param name="selector">Selector function that returns true if the data of a given vertex belongs to set X, or false otherwise.</param>
        public Boolean IsBipartite(Func<V, Boolean> selector)
        {
            foreach (Edge e in edges)
                if ((selector(e.Start.Data) && selector(e.End.Data)) || (!selector(e.End.Data) && !selector(e.Start.Data)))
                    return false;
            return true;
        }

        /// <summary>
        /// Check whether this graph is acyclic, i.e. there are no strongly connected components but its
        /// simple vertices. To detected strongly connected components, a default implementation of the
        /// Tarjan algorithm is used.
        /// </summary>
        public Boolean IsAcyclic()
        {
            if (defaultSccProvider == null)
                defaultSccProvider = TarjanAlgorithmProvider.Create(this);
            return this.IsAcyclic(defaultSccProvider);
        }

        /// <summary>
        /// Check whether this graph is acyclic, i.e. there are no strongly connected components but its
        /// simple vertices. To detected strongly connected components, the method uses the given algorithm.
        /// </summary>
        /// <param name="sccProvider">The algorithm provider to use when calculating the strongly connected components.</param>
        public Boolean IsAcyclic(IStronglyConnectedComponentsProvider<V, E> sccProvider)
        {
            var scComponents = sccProvider.Compute();
            return scComponents.All(scc => scc.Count == 1);
        }
    }

    /// <summary>
    /// Represents a weighted graph, i.e. a colored graph where the color function for the edges returns
    /// a numeric value.
    /// </summary>
    /// <typeparam name="V">This is the type of data contained in the graph's vertices. It must implement IEquatable<V>.</typeparam>
    /// <typeparam name="E">This is the type of data associated to the graph's edges. It must implement IComparable<E>.</typeparam>
    public class WeightedGraph<V, E> : Graph<V, E>
        where V : IEquatable<V>
        where E : IEquatable<E>
    {
        /// <summary>
        /// Represents a walk on a graph.
        /// </summary>
        public class Walk : ReadOnlyCollection<Edge>
        {
            /// <summary>
            /// Indicates if the walk is closed, i.e. the first and last vertices are equals.
            /// </summary>
            public Boolean IsClosed { get; private set; }
            /// <summary>
            /// Indicates if the walk is a cycle.
            /// </summary>
            public Boolean IsCycle { get; private set; }
            /// <summary>
            /// Indicates if the walk is a path.
            /// </summary>
            public Boolean IsPath { get; private set; }
            /// <summary>
            /// Indicates if a path is trivial.
            /// </summary>
            public Boolean IsTrivialPath { get; private set; }

            /// <summary>
            /// Retrieves a collection of all vertices in the walk.
            /// </summary>
            public IEnumerable<Vertex> Vertices
            {
                get
                {
                    foreach (Edge e in this)
                        yield return e.Start;
                    if (this.Count > 0)
                        yield return this.Last().End;
                }
            }

            /// <summary>
            /// Creates a walk from a succession of adjacent edges.
            /// </summary>
            public Walk(IList<Edge> ed) : base(ed)
            {
                if (ed.Count > 0)
                {
                    this.IsClosed = ed.First().Equals(ed.Last());

                    var vertices = ed.Select(e => e.Start).ToList();
                    vertices.Add(ed.Last().End);

                    this.IsPath = true;
                    this.IsCycle = true;
                    for (Int32 i = 0; i < vertices.Count && (this.IsPath || this.IsCycle); i++)
                        for (Int32 j = i + 1; j < vertices.Count; j++)
                            if ((i != j) && (vertices[i].Equals(vertices[j])))
                            {
                                this.IsPath = false;
                                if (!((i == 0 && j == vertices.Count - 1) || (j == 0 && i == vertices.Count - 1)))
                                    this.IsCycle = false;

                                if (!this.IsPath && !this.IsCycle)
                                    break;
                            }
                    this.IsCycle &= this.IsClosed;
                    this.IsTrivialPath = false;
                }
                else
                {
                    this.IsClosed = this.IsPath = this.IsCycle = this.IsTrivialPath = true;
                }
            }

            /// <summary>
            /// Check whether two walks intersect.
            /// </summary>
            public Boolean Intersects(Walk otherWalk)
            {
                return this.Vertices.Intersect(otherWalk.Vertices).Count() > 0;
            }

            /// <summary>
            /// Check whether two walks are independent, i.e. if they have no vertices in common except, at most, the end.
            /// </summary>
            public Boolean IsIndependentOf(Walk otherWalk)
            {
                var commonVertices = this.Vertices.Intersect(otherWalk.Vertices);

                return (commonVertices.Count() == 0) || (commonVertices.Contains(this.Last().End));
            }
        }

        protected IAllPairsShortestPathsProvider<V, E> defaultAllPairsProvider;

        /// <summary>
        /// Contains an anonymous function that computes the weight of an edge based on its data.
        /// </summary>
        public Func<Edge, Double> EdgeWeight { get; protected set; }

        /// <summary>
        /// Initializes a new empty weighted graph specifying a new weight function.
        /// </summary>
        /// <param name="weightFunction">A weight function that, given the edge data, calculates its weight as a floating point value.</param>
        public WeightedGraph(Func<E, Double> weightFunction)
        {
            this.EdgeWeight = edge => weightFunction(edge.Data);
        }

        /// <summary>
        /// Check whether the graph is connected. In order to obtain the result, all the shortest paths are
        /// computed for every pair of vertices using a default Floyd-Warshall algorithm.
        /// </summary>
        public Boolean IsConnected()
        {
            if (defaultAllPairsProvider == null)
                defaultAllPairsProvider = FloydWarshallAlgorithmProvider.Create(this);

            return this.IsConnected(defaultAllPairsProvider);
        }

        /// <summary>
        /// Check whether the graph is connected. In order to obtain the result, all the shortest paths are
        /// computed for every pair of vertices using the specified algorithm.
        /// </summary>
        /// <param name="allPairsProvider">The algorithm provider to use when calcolating all the shortest paths pairs.</param>
        public Boolean IsConnected(IAllPairsShortestPathsProvider<V, E> allPairsProvider)
        {
            Walk[,] shortestPaths = allPairsProvider.Compute();
            for (Int32 i = 0; i < this.Order; i++)
                for (Int32 j = 0; j < this.Order; j++)
                    if (shortestPaths[i, j] == null)
                        return false;
            return true;
        }
    }
}
