﻿using System;
using System.Collections.Generic;

namespace Hamlet
{
    /// <summary>
    /// Represents an implementation of a strongly connected components detection algorithm.
    /// </summary>
    /// <typeparam name="V"></typeparam>
    /// <typeparam name="E"></typeparam>
    public interface IStronglyConnectedComponentsProvider<V, E>
        where V : IEquatable<V>
        where E : IEquatable<E>
    {
        IEnumerable<Graph<V, E>.StronglyConnectedComponent> Compute();
    }
    
    /// <summary>
    /// Implements the Tarjan algorithm for strongly connected components detection on directed graphs.
    /// </summary>
    public class TarjanAlgorithmProvider
    {
        public static TarjanAlgorithmProvider<V, E> Create<V, E>(Graph<V, E> graph)
            where V : IEquatable<V>
            where E : IEquatable<E>
        {
            return new TarjanAlgorithmProvider<V, E>(graph);
        }
    }

    /// <summary>
    /// Implements the Tarjan algorithm for strongly connected components detection on directed graphs.
    /// </summary>
    public class TarjanAlgorithmProvider<V, E> : TarjanAlgorithmProvider, IStronglyConnectedComponentsProvider<V, E>
        where V : IEquatable<V>
        where E : IEquatable<E>
    {
        private class VertexData
        {
            public Int32 Index { get; set; }
            public Int32 LowLink { get; set; }
        }

        private Dictionary<Graph<V, E>.Vertex, VertexData> data;
        private Int32 index;
        private Stack<Graph<V, E>.Vertex> stack;
        private List<Graph<V, E>.StronglyConnectedComponent> results;
        protected Graph<V, E> graph;

        public TarjanAlgorithmProvider(Graph<V, E> graph)
        {
            this.graph = graph;
            data = new Dictionary<Graph<V, E>.Vertex, VertexData>();
            stack = new Stack<Graph<V, E>.Vertex>();
            results = new List<Graph<V, E>.StronglyConnectedComponent>();
        }

        public IEnumerable<Graph<V, E>.StronglyConnectedComponent> Compute()
        {
            if (!graph.IsDirected)
                throw new InvalidOperationException("Tarjan's algorithm only work for directed graphs.");

            this.InitializeData();

            foreach (var v in graph.Vertices)
                if (data[v].Index < 0)
                    this.StrongConnect(v);

            this.DisposeData();

            return results;
        }

        private void InitializeData()
        {
            index = 0;
            results.Clear();
            foreach (var v in graph.Vertices)
                data.Add(v, new VertexData() { Index = -1, LowLink = -1 });
        }

        private void DisposeData()
        {
            data.Clear();
            stack.Clear();
        }

        private void StrongConnect(Graph<V, E>.Vertex v)
        {
            data[v].Index = index;
            data[v].LowLink = index;
            index++;
            stack.Push(v);

            foreach (var vw in graph.FindEdges(v, EdgeEnds.Start))
            {
                var w = vw.End;
                if (data[w].Index < 0)
                {
                    this.StrongConnect(w);
                    data[v].LowLink = Math.Min(data[v].LowLink, data[w].LowLink);
                }
                else if (stack.Contains(w))
                    data[v].LowLink = Math.Min(data[v].LowLink, data[w].Index);
            }

            if (data[v].LowLink == data[v].Index)
            {
                // A new strongly connected component
                List<Graph<V, E>.Vertex> scc = new List<Graph<V, E>.Vertex>();
                Graph<V, E>.Vertex w;
                do
                {
                    w = stack.Pop();
                    scc.Add(w);
                } while (!w.Equals(v));

                results.Add(new Graph<V, E>.StronglyConnectedComponent(scc));
            }
        }
    }
}
